# -*- coding: utf-8 -*-
"""
Created on Thu Dec 19 12:07:05 2019

@author: Angel Garcia Olaya PLG-UC3M
@version: 1.0
A representation of the Date class for problem 2 of the extra exercises.
 Notice the problem can be also solved by using functions, with no OOP.
"""


class Date:

    def __init__(self, form):
        ''' To force the user to select one of the formats, the only way to
        create objects is self'''
        # Private constants to handle date formats easily
        self.__DAY_FIRST = 1
        self.__MONTH_FIRST = 2
        self.__YEAR_FIRST = 3
        self.__TEXT = 4
        # If the format is not valid, we choose DAY_FIRST
        if (form < 0 or form > 4):
            self.format = self.__DAY_FIRST
        else:
            self.format = form
        # Private dictionary to set the names and days of each month
        self.__month_days = {'january': 31, 'february': 28, 'march': 31,
                             'april': 30,
                             'may': 31, 'june': 30, 'july': 31, 'august': 31,
                             'september': 30, 'october': 31, 'november': 30,
                             'december': 31}
        # These will be properties, despite the format we store data as numbers
        self.year = 1900
        self.month = 1
        self.day = 1

    @property
    def year(self) -> int:
        ''' This method declares the year property, now year is not an
        attribute but a property. Externally there is no difference between
        both '''
        # Notice that in the property method, we treat year as private even if
        # it is public
        return self.__year

    # The name of this method must be the same as the name of the attribute,
    # too. The decoration must be @attribute_name.setter
    @year.setter
    def year(self, year: int):
        ''' This method is used to change the value of the year, if it is not
        zero it will take it, if not it will set the year to 1900. It also
        sets the values of month_days and leap_year fields as both depend
        on the year'''
        if year != 0:
            # Notice that we are defining all the attributes as private
            self.__year = year
        else:
            self.__year = 1900
        # We calculate the value of leap year
        self.__leap_year = self.__year % 4 == 0 and (self.__year % 100 != 0 or
                                                     self.__year % 400 == 0)
        # If we are in a leap year, february has 29 days
        if self.__leap_year:
            self.__month_days['february'] = 29
        else:
            self.__month_days['february'] = 28

    @property
    def month(self) -> int:
        ''' This method declares the month property, now month is not an
        attribute but a property. Externally there is no difference between
        both '''
        return self.__month

    @month.setter
    def month(self, month: int):
        ''' Method to change the value of the month. It controls the month
        is correct '''
        # Checking the month is correct, else we don't change it
        if month > 0 and month < 13:
            self.__month = month

    @property
    def day(self):
        ''' This method declares the day property, now day is not an attribute
        but a property. Externally there is no difference between both '''
        return self.__day

    @day.setter
    def day(self, d):
        ''' This method is the "day property" setter, it allows us to control
        the valid values for day'''
        # Same for day
        if d >= 1 and d <= self.days_in_month():
            self.__day = d
        else:
            self.__day = 1

    @property
    def leap_year(self):
        ''' Returns the value of the private field leap_year. Using this we
        allow to read its value, but as we are not providing setter method,
        it cannot be changed (it only changes when year changes)'''
        return self.__leap_year

    def month_name(self):
        ''' Returns the name of the month '''
        # To get the name of the month, we convert the dictionary to a tuple
        # this will create a tuple with just the keys
        month_name = tuple(self.__month_days)[self.month-1]
        return month_name

    def days_in_month(self):
        ''' Returns the number of days of this month'''
        return self.__month_days[self.month_name()]

    def setDate(self, date):
        '''Sets the date receiving a string in any of the four formats. If
        either month, day or year are not correct they remain unchanged '''
        # Invoking another method to extract day, month and year
        params = self.extractParameters(date)
        self.year = int(params[0])
        self.month = int(params[1])
        self.day = int(params[2])

    def extractParameters(self, d) -> list:
        ''' Extracts the year, month and day from a  and returns a list with
            their numeric values. We assume that the user will enter
            numbers where numbers are supposed '''
        if (self.format == self.__TEXT):
            # We are assuming there will be 3 items separated by a space, if
            # not a runtime error will appear
            result = []
            parts = d.split(" ")
            # Year is the last element
            result.append(int(parts[2]))
            # To get the number of the month, we convert the dictionary to a
            # tuple, this will create a tuple with just the keys
            month_number = tuple(self.__month_days).index(parts[0].lower())+1
            result.append(month_number)
            # Day is the second, but extra stuff needs to be removed
            # actually the last 3 characters need to be removed
            result.append(int(parts[1][0:-3]))

        else:
            # We are assuming there will be 3 items separated by "/", if not a
            # runtime error will appear
            parts = d.split("/")
            if self.format == self.__DAY_FIRST:
                result = [parts[2], parts[1], parts[0]]
            elif self.format == self.__MONTH_FIRST:
                result = [parts[2], parts[0], parts[1]]
            else:
                result = [parts[0], parts[1], parts[2]]
        return result

    def setFormat(self, form):
        ''' Converts the date to the format (changes the format field, the
        inner representation of the Date does not change). If the format is not
        correct it does nothing'''
        if (form >= self.__DAY_FIRST and form <= self.__TEXT):
            self.format = form

    def __str__(self):
        ''' Returns a string representing the date in the desired format'''
        def ordinalMark(n):
            ''' Returns the ordinal mark for a given number (st,nd,rd,th) '''
            if (n > 3 and n < 21):
                return "th"
            # Converting it to String to make it easier
            number = str(n)
            if number[-1] == '1':
                # works for 1, 21 and 31
                return "st"
            elif number[-1] == '2':
                return "nd"
            elif number[-1] == '3':
                return "rd"
            else:
                return "th"

        if (self.day < 10):
            day = "0" + str(self.day)
        else:
            day = str(self.day)
        # To get the number of the month, we convert the dictionary to a tuple
        # this will create a tuple with just the keys
        if (self.month < 9):
            month = "0" + str(self.month)
        else:
            month = str(self.month)

        if self.format == self.__DAY_FIRST:
            return day + "/" + month + "/" + str(self.year)
        elif self.format == self.__MONTH_FIRST:
            return month + "/" + day + "/" + str(self.year)
        elif self.format == self.__YEAR_FIRST:
            return str(self.year) + "/" + month + "/" + day
        else:
            return (self.month_name() + " " + str(self.day)
                    + ordinalMark(self.day) + ", " + str(self.year))

    def __repr__(self):
        return self.__str__()

    def nextDay(self):
        ''' Returns the next day after the current one'''
        # Creating a new Date object with the data of this one
        nxt = Date(self.format)
        nxt.day = self.day
        nxt.month = self.month
        nxt.year = self.year
        if nxt.day < nxt.days_in_month():
            nxt.day += 1
        else:
            nxt.day = 1
            if nxt.month < 12:
                nxt.month += 1
            else:
                nxt.month = 1
                nxt.year += 1
        return nxt

    def isPrevious(self, d):
        '''Returns true is this date is previous to the parameter one'''
        return (self.year < d.year
                or (self.year == d.year
                    and (self.month < d.month
                         or (self.month == d.month and self.day < d.day))))

    def countYears(self, other):
        ''' Counts the years between this date and the parameter.'''
        return abs(self.year-other.year)

    def __days_to_date(self, other):
        ''' Private method to chek the days between to dates of the same year.
        The second will always be >= than the first'''
        # If they belong to the same month
        if self.month == other.month:
            return other.day - self.day
        else:
            days = 0
            # Till the end of the month
            days += self.days_in_month()-self.day
            # From the beginning of the second month
            days += other.day
            # Days in between
            month_days = tuple(self.__month_days.values())
            for m in range(self.month + 1, other.month):
                days += month_days[m]
        return days

    def countDays(self, d):
        ''' Counts the days between two dates'''
        # We put first the oldest date
        if self.isPrevious(d):
            first = self
            second = d
        else:
            first = d
            second = self
        # If they belong to the same year, we just count the days using the
        # previous method
        if (first.year == second.year):
            return first.__days_to_date(second)
        # If not, we will count first the days to the end of the year for the
        # earliest date of both, plus the days from the beginning of the year
        # for the latest date of both
        # Days from the first date until the end of the year
        aux = Date(self.__DAY_FIRST)
        aux.setDate('31/12/'+str(self.year))
        days = first.__days_to_date(aux) + 1
        # Days from the beginning of the year until second object's date
        aux.setDate('1/1/'+str(d.year))
        days += aux.__days_to_date(second)
        # We counted the days of first.year until the end of the year, so we
        # start in the next year
        startingYear = first.year + 1
        endYear = second.year
        # Now we count days in whole years
        for ii in range(startingYear, endYear):
            if ii % 4 == 0 and (ii % 100 != 0 or ii % 400 == 0):
                days += 366
            else:
                days += 365
        return days

    def print_days(self, d):
        '''Returns a  with all the days between self and the parameter'''
        res = []
        # We put first the oldest date
        if self.isPrevious(d):
            first = self
            second = d
        else:
            first = d
            second = self
        # And now append elements to the list
        while first.isPrevious(second):
            first = first.nextDay()
            res.append(first)
        return res

    def in_Between(self, d1, d2):
        '''Returns true if self date is between the two ones passed as
        parameters'''
        if d1.isPrevious(self) and self.isPrevious(d2):
            return True
        else:
            return False

    def sort_dates(self, aList):
        ''' This function sorts a list of dates using bubble algorithm'''
        # Variable to stop working if the list is already sorted
        swapping = True
        # number of iterations of the outer loop
        num = len(aList) - 1
        # outer loop: (elements-1) loops max
        while num > 0 and swapping:
            # At the beginning of each iteration we haven't swapped
            swapping = False
            # Inner loop (len(list)-number comparisons)
            for i in range(num):
                # if the lower index element is bigger than the next one, swap
                if aList[i + 1].isPrevious(aList[i]):
                    # we did a change, so we update the variable
                    swapping = True
                    # we use aux to swap, notice that in Python we don't really
                    # need this aux variable as we can do:
                    # aList[i], aList[i + 1] = aList[i + 1], aList[i]
                    aux = aList[i]
                    aList[i] = aList[i + 1]
                    aList[i + 1] = aux
            # One iteration performed, we decrease the number of pending ones
            num -= 1
