#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Nov 22 11:44:56 2018

@author: yolanda
"""

import measure


class Station:

    def __init__(self, idStation, location, hour):
        self.idStation = idStation
        self.location = location
        if hour > 0:
            self.measurePerHour = hour
        else:
            self.measurePerHour = 1
        self.temperatures = []
        self.pressures = []
        self.meanTemperature = 0.0
        self.meanPressure = 0.0
        self.maxMeasures = 24 * self.measurePerHour

    def createMeasure(self, measureType):
        if measureType:
            if len(self.temperatures) < self.maxMeasures:
                measureT = measure.Temperature()
                self.temperatures.append(measureT)
                return True
            else:
                return False
        else:
            if len(self.pressures) < self.maxMeasures:
                measureP = measure.Pressure()
                self.pressures.append(measureP)
                return True
            else:
                return False

    def addMeasures(self, measureType, number):
        r, i = True, 0
        while r and i < number:
            r = self.createMeasure(measureType)
            i += 1
        self.calculateMean(measureType)
        return r

    def calculateMean(self, measureType):
        if measureType and len(self.temperatures) > 0:
            meanT = 0
            for t in self.temperatures:
                meanT += t.temperature
            self.meanTemperature = meanT / len(self.temperatures)
        elif len(self.pressures) > 0:
            meanP = 0
            for p in self.pressures:
                meanP += p.pressure
            self.meanPreassure = meanP / len(self.pressures)

    def sortMeasure(self, aList):
        for i in range(len(aList)):
            measure = aList[i]
            j = i - 1
            while j >= 0 and measure.timestamp.before(aList[j].timestamp):
                aList[j + 1] = list[j]
                j -= 1
            aList[j + 1] = measure

    def readTemperatureTS(self, timeStamp):
        if len(self.temperatures) == 0:
            return
        closest = abs(self.temperatures[0].timestamp.secondsTo(timeStamp))
        found = False
        pos, i = 0, 0
        while not found and i < len(self.temperatures):
            current = abs(self.temperatures[i].timestamp.secondsTo(timeStamp))
            if current <= closest:
                closest = current
                read = self.temperatures[i]
                pos = i
            else:
                found = True
            i += 1
        del self.temperatures[pos]
        self.calculateMean(True)
        return read

    def readTemperatureP(self, pos):
        if pos >= 0 and pos < len(self.temperatures):
            read = self.temperatures[pos]
            del self.temperatures[pos]
            self.calculateMean(True)
            return read

    def readTemperature(self):
        read = []
        for t in self.temperatures:
            read.append(t)
        self.temperatures.clear()
        self.calculateMean(True)
        return read

    def readPressureTS(self, timeStamp):
        if len(self.pressures) == 0:
            return
        closest = abs(self.pressures[0].timestamp.secondsTo(timeStamp))
        found = False
        pos, i = 0, 0
        while not found and i < len(self.pressures):
            current = abs(self.pressures[i].timestamp.secondsTo(timeStamp))
            if current <= closest:
                closest = current
                read = self.pressures[i]
                pos = i
            else:
                found = True
            i += 1
        del self.pressures[pos]
        self.calculateMean(False)
        return read

    def readPressureP(self, pos):
        if pos >= 0 and pos < len(self.pressures):
            read = self.pressures[pos]
            del self.pressures[pos]
            self.calculateMean(False)
            return read

    def readPressure(self,):
        read = []
        for p in self.pressures:
            read.append(p)
        self.pressures.clear()
        self.calculateMean(False)
        return read

    def __str__(self,):
        s = 'Temperatures:\n'
        for t in self.temperatures:
            s += str(t.temperature) + '\n'
        s += 'Pressures:\n'
        for p in self.pressures:
            s += str(p.pressure) + '\n'
        return s
